# coding: utf-8


def format_error(error):
    if 'publicMessage' in error and len(error) == 1:
        return error['publicMessage']
    elif 'publicMessage' not in error:
        return 'Error code: %s. %s' % (error['err_code'], error['err_msg'])
    return 'Error code: %s. %s %s' % (error['err_code'], error['publicMessage'], error['err_msg'])


class HasOffersException(BaseException):

    def __str__(self):
        return '\n'.join(format_error(error) for error in self.errors)

    @property
    def errors(self):
        value = self.args[0]
        if not isinstance(value, list):
            value = [value]
        return value


class MaxRetriesExceeded(HasOffersException):

    def __str__(self):
        return 'Max retries exceeded'



# coding: utf-8
from .core import Model, ModelManager


class Advertiser(Model):
    """
    An Advertiser account.
    """
    generic_methods = ('update',)

    def block(self, reason=''):
        return self._manager.block(self.id, reason)


class AdvertiserManager(ModelManager):
    model = Advertiser
    name = 'advertisers'
    generic_methods = (
        'create',
        'update',
        'find_by_id',
        'find_all',
        'find_all_ids',
    )

    def block(self, id, reason=''):
        return self._call('block', id=id, reason=reason)



# coding: utf-8
from .advertiser import Advertiser, AdvertiserManager
from .conversion import Conversion, ConversionManager
from .core import ModelManager
from .country import Country, CountryManager
from .goal import Goal, GoalManager
from .offer import Offer, OfferManager
from .raw_log import RawLogManager


MODEL_MANAGERS = (
    AdvertiserManager, ConversionManager, CountryManager, GoalManager, OfferManager, RawLogManager
)



# coding: utf-8
from csv import DictReader
from io import BytesIO, StringIO
from zipfile import ZipFile

from pip.utils import cached_property

from .core import Model, ModelManager


class LogItem(Model):
    """
    Log entities have no id field, because they should be compared in a different way.
    """

    def __eq__(self, other):
        return isinstance(other, self.__class__) and self._data == other._data


class DateDir(LogItem):
    """
    Directory with log files.
    """

    def __str__(self):
        return '%s: %s (%s)' % (self.__class__.__name__, self.displayName, self.dirName)

    def list_logs(self):
        return self._manager.list_logs(dir_name=self.dirName)


class LogFile(LogItem):
    """
    Log file.
    """

    def __str__(self):
        return '%s: %s (%s)' % (self.__class__.__name__, self.displayName, self.filename)

    @cached_property
    def download_link(self):
        return self._manager.get_download_link(log_filename=self.filename)

    @cached_property
    def content(self):
        """
        Returns raw CSV content of the log file.
        """
        raw_content = self._manager.api.session.get(self.download_link).content
        data = BytesIO(raw_content)
        archive = ZipFile(data)
        filename = archive.filelist[0]  # Always 1 file in the archive
        return archive.read(filename)

    @cached_property
    def records(self):
        data = StringIO(self.content.decode())
        return [LogRecord(manager=self._manager, **kwargs) for kwargs in DictReader(data)]


class LogRecord(LogItem):
    """
    Log record with all data about clicks / conversions / impressions.
    """

    def __str__(self):
        return '%s: %s (%s)' % (self.__class__.__name__, self.offer_id, self.transaction_id)


class LogItemManager(ModelManager):
    """
    Works with exact type of log.
    """

    def __init__(self, *args, log_type, **kwargs):
        self.log_type = log_type
        super().__init__(*args, **kwargs)

    def _call(self, method, **kwargs):
        return self.api._call('RawLog', method, log_type=self.log_type, **kwargs)

    def list_date_dirs(self):
        response = self._call('listDateDirs', raw=True)
        return [DateDir(manager=self, **kwargs) for kwargs in response['dateDirs']]

    def list_logs(self, dir_name):
        response = self._call('listLogs', date_dir=dir_name, raw=True)
        return [LogFile(manager=self, **kwargs) for kwargs in response['logFiles']]

    def get_download_link(self, log_filename):
        return self._call('getDownloadLink', log_filename=log_filename, raw=True)['link']

    def find_all(self, date):
        return sum([log_file.records for log_file in self.list_logs(date)], [])


class RawLogManager(ModelManager):
    """
    Works as proxy for other managers.
    """
    name = 'raw_logs'
    log_types = (
        'clicks',
        'conversions',
        'impressions',
    )

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.setup_managers()

    def setup_managers(self):
        for log_type in self.log_types:
            setattr(self, log_type, LogItemManager(api=self.api, log_type=log_type))



# coding: utf-8
from collections import OrderedDict

import requests
import time

from .exceptions import HasOffersException, MaxRetriesExceeded
from .logging import get_logger
from .models import MODEL_MANAGERS
from .utils import prepare_query_params


def is_empty(data):
    return isinstance(data, bool) or not data


def is_paginated(data):
    return 'pageCount' in data


def retry(method):
    """
    Allows to retry method execution few times.
    """

    def inner(self, *args, **kwargs):
        attempt_number = 1
        while attempt_number < self.retries:
            try:
                return method(self, *args, **kwargs)
            except HasOffersException as exc:
                if not 'API usage exceeded rate limit' in str(exc):
                    raise exc
                self.logger.debug('Retrying due: %s', exc)
            time.sleep(self.retry_timeout)
            attempt_number += 1
        raise MaxRetriesExceeded

    return inner


class HasOffersAPI:
    """
    Client to communicate with HasOffers API.
    """

    def __init__(self, endpoint=None, network_token=None, network_id=None, verify=True, retries=3, retry_timeout=3,
                 verbosity=0):
        self.endpoint = endpoint
        self.network_token = network_token
        self.network_id = network_id
        self.verify = verify
        self.retries = retries
        self.retry_timeout = retry_timeout
        self.logger = get_logger(verbosity)
        self.setup_managers()

    def setup_managers(self):
        """
        Allows to access manager by model name - it is convenient, because HasOffers returns model names in responses.
        """
        self._managers = {}
        for manager_class in MODEL_MANAGERS:
            instance = manager_class(self)
            setattr(self, instance.name, instance)
            if instance.model:
                self._managers[instance.model.__name__] = instance

    def __str__(self):
        return '%s: %s / %s' % (self.__class__.__name__, self.network_token, self.network_id)

    def __repr__(self):
        return '<%s>' % self

    @property
    def session(self):
        if not hasattr(self, '_session'):
            self._session = requests.Session()
        return self._session

    @retry
    def _call(self, target, method, single_result=True, raw=False, **kwargs):
        """
        Low-level call to HasOffers API.
        """
        params = prepare_query_params(
            NetworkToken=self.network_token,
            NetworkId=self.network_id,
            Target=target,
            Method=method,
            **kwargs
        )
        self.logger.debug('Request parameters: %s', params)
        response = self.session.get(self.endpoint, params=params, verify=self.verify)
        self.logger.debug('Response [%s]: %s', response.status_code, response.text)
        response.raise_for_status()
        data = response.json(object_pairs_hook=OrderedDict)
        return self.handle_response(data, target=target, single_result=single_result, raw=raw)

    def handle_response(self, content, target=None, single_result=True, raw=False):
        """
        Parses response, checks it.
        """
        response = content['response']

        self.check_errors(response)

        data = response.get('data')

        if is_empty(data):
            return data
        elif is_paginated(data):
            if not data['count']:
                return data['data']
            data = data['data']

        if raw:
            return data
        return self.init_all_objects(data, target=target, single_result=single_result)

    def check_errors(self, response):
        errors = response.get('errors')
        if errors:
            raise HasOffersException(errors)

    def init_all_objects(self, data, target=None, single_result=True):
        """
        Initializes model instances from given data.
        Returns single instance if single_result=True.
        """
        if single_result:
            return self.init_target_object(target, data)
        return list(self.expand_models(target, data))

    def init_target_object(self, target, data):
        """
        Initializes target object and assign extra objects to target as attributes
        """
        target_object = self.init_single_object(target, data.pop(target))
        for key, item in data.items():
            setattr(target_object, key.lower(), self.init_single_object(key, item))
        return target_object

    def init_single_object(self, target, data):
        if data:
            return self._managers[target].init_instance(data)

    def expand_models(self, target, data):
        """
        Generates all objects from given data.
        """
        if isinstance(data, dict):
            data = data.values()
        for chunk in data:
            if target in chunk:
                yield self.init_target_object(target, chunk)
            else:
                for key, item in chunk.items():
                    yield self.init_single_object(key, item)



# coding: utf-8
from .conversion import ConversionManager
from .core import Model, ModelManager, RelatedManager


class RelatedConversionsManager(RelatedManager):
    related_object_name = 'offer_id'


class Offer(Model):
    """
    An Offer.
    """
    generic_methods = ('update',)

    @property
    def conversions(self):
        return RelatedConversionsManager(api=self._manager.api, base_manager_class=ConversionManager, id=self.id)

    def add_target_country(self, country_code):
        return self._manager.add_target_country(self.id, country_code)

    def get_target_countries(self):
        return self._manager.get_target_countries(self.id)

    def add_category(self, category_id):
        return self._manager.add_category(self.id, category_id)


class OfferManager(ModelManager):
    model = Offer
    name = 'offers'
    generic_methods = (
        'create',
        'update',
        'find_by_id',
        'find_all',
        'find_all_ids',
    )

    def add_target_country(self, id, country_code):
        return self._call('addTargetCountry', id=id, country_code=country_code)

    def get_target_countries(self, id):
        return self._call('getTargetCountries', id=id, single_result=False)

    def add_category(self, id, category_id):
        return self._call('addCategory', id=id, category_id=category_id)



# coding: utf-8
from .core import Model, ModelManager


class Conversion(Model):
    """
    A Conversion.
    """
    generic_methods = ('update', )


class ConversionManager(ModelManager):
    model = Conversion
    name = 'conversions'
    generic_methods = (
        'create',
        'update',
        'find_by_id',
        'find_all',
    )



# coding: utf-8
from collections import defaultdict

from ..utils import Filter, Sort


class SelectiveInheritanceMeta(type):
    """
    Allows to inherit only methods from `generic_methods` attribute.
    """
    generic_methods = defaultdict(dict)

    def __new__(mcs, name, bases, members):
        if not bases:
            for key, value in list(members.items()):
                if getattr(value, '_is_generic', None):
                    mcs.generic_methods[name][key] = members.pop(key)
        else:
            for method in members.get('generic_methods', ()):
                members[method] = mcs.generic_methods[bases[0].__name__][method]
        return super().__new__(mcs, name, bases, members)


def generic_method(method):
    """
    Marks method as generic.
    """
    method._is_generic = True
    return method


class Model(metaclass=SelectiveInheritanceMeta):
    """
    Abstract model for HasOffers entity.
    """
    generic_methods = ()

    def __init__(self, manager, **kwargs):
        self._manager = manager
        self._data = kwargs
        self.__dict__.update(kwargs)

    def __str__(self):
        return '%s: %s' % (self.__class__.__name__, self._data.get('id'))

    def __repr__(self):
        return '<%s>' % self

    def __eq__(self, other):
        return isinstance(other, self.__class__) and self.id == other.id

    def as_dict(self):
        return self._data

    @generic_method
    def update(self, **kwargs):
        return self._manager.update(self.id, **kwargs)


class ModelManager(metaclass=SelectiveInheritanceMeta):
    """
    Proxy for API methods with predefined model.
    Used to structure API. Every manager works only with single model methods.
    """
    model = None
    name = None
    generic_methods = ()

    def __init__(self, api):
        self.api = api

    def init_instance(self, data):
        return self.model(manager=self, **data)

    def _call(self, method, **kwargs):
        return self.api._call(self.model.__name__, method, **kwargs)

    @generic_method
    def create(self, **kwargs):
        return self._call('create', data=kwargs)

    @generic_method
    def update(self, id, **kwargs):
        return self._call('update', id=id, data=kwargs)

    @generic_method
    def find_by_id(self, id, fields=None, contain=None):
        assert fields is None or isinstance(fields, (tuple, list)), 'Fields should be a tuple or list'
        return self._call('findById', id=id, fields=fields, contain=contain)

    @generic_method
    def find_all(self, sort=(), limit=None, page=None, fields=None, contain=None, **kwargs):
        assert limit is None or isinstance(limit, int), 'Limit should be an integer'
        assert page is None or isinstance(page, int), 'Page should be an integer'
        assert fields is None or isinstance(fields, (tuple, list)), 'Fields should be a tuple or list'
        return self._call(
            'findAll',
            filters=Filter(**kwargs),
            sort=Sort(sort, self.model.__name__),
            limit=limit, page=page, fields=fields, contain=contain, single_result=False
        )

    @generic_method
    def find_all_ids(self, **kwargs):
        return self._call('findAllIds', filters=Filter(**kwargs), single_result=False, raw=True)


class RelatedManager:
    related_object_name = None

    def __init__(self, api, base_manager_class, id):
        self.base_manager = base_manager_class(api)
        self.id = id

    def find_all(self, **kwargs):
        kwargs[self.related_object_name] = self.id
        return self.base_manager.find_all(**kwargs)
